import matplotlib.pyplot as plt
from matplotlib.patches import Patch, Rectangle
from mpl_toolkits.axes_grid1 import AxesGrid
from mpl_toolkits.mplot3d import Axes3D
import numpy as np


def show_overview(d, cell_to_preview, gain_to_preview, out_folder=None, infix=None):
    """
    Show an overview
    :param d: A dict with the number of modules and
        a dict for the constant names and it's data.
    :param cell_to_preview: Integer number of memory cells to preview.
    :param gain_to_preview: Integer number for the gain stages to preview.
    :param out_folder: Output folder for saving the plotted.png image.
    :param infix: infix to include in the png file name.
    :return:
    """
    for module, data in d.items():
        # Adapt number of columns to number of constants in d.
        ncols = (len(d[module]) + (2 - 1)) // 2
        fig = plt.figure(figsize=(20, 20))
        grid = AxesGrid(fig, 111,
                        nrows_ncols=(2, ncols),
                        axes_pad=(0.9, 0.15),
                        label_mode="1",
                        share_all=True,
                        cbar_location="right",
                        cbar_mode="each",
                        cbar_size="7%",
                        cbar_pad="2%",
                        )
        i = 0
        for key, item in data.items():
            cf = 0
            if "ThresholdsDark" in key:
                cf = -1
            if len(item.shape) == 4:
                med = np.nanmedian(item[..., cell_to_preview,
                                        gain_to_preview + cf])
            else:
                med = np.nanmedian(item[..., cell_to_preview])

            bound = 0.2
            while (np.count_nonzero((item[..., cell_to_preview, gain_to_preview + cf] < med - np.abs(bound * med)) |  # noqa
                                    (item[..., cell_to_preview, gain_to_preview + cf] > med + np.abs(bound * med))) /  # noqa
                   item[..., cell_to_preview, gain_to_preview + cf].size > 0.01):  # noqa
                bound *= 2

            if "BadPixelsDark" in key:
                im = grid[i].imshow(np.log2(item[..., cell_to_preview,
                                                 gain_to_preview + cf]),
                                    interpolation="nearest", vmin=0, vmax=8,
                                    aspect='auto')
            else:

                if len(item.shape) == 4:
                    im_prev = item[..., cell_to_preview, gain_to_preview + cf]
                    vmax = np.abs(med + bound * med)
                else:
                    im_prev = item[..., cell_to_preview]
                    # move the axis of the image to show horizontally
                    # on the output report.
                    if im_prev.shape[0] > im_prev.shape[1]:
                        im_prev = np.moveaxis(item[..., cell_to_preview], 0, 1)
                    vmax = med + np.abs(bound * med)

                im = grid[i].imshow(im_prev, interpolation="nearest",
                                    vmin=med - np.abs(bound * med),
                                    vmax=vmax, aspect='auto')

            cb = grid.cbar_axes[i].colorbar(im)
            cb.set_label_text("ADU" if key != "BadPixels" else "Bad Pixel Code")

            grid[i].text(5, 20, key, color="w" if key != "BadPixels" else "k", fontsize=20)  # noqa

            i += 1
        grid[0].text(5, 50, module, color="r" if key != "BadPixels" else "k", fontsize=20)  # noqa
        if out_folder and infix:
            fig.savefig(f"{out_folder}/"
                        f"dark_analysis_{infix}_module_{module}.png")


def rebin(a, *args):
    '''rebin ndarray data into a smaller ndarray of the same rank whose dimensions
    are factors of the original dimensions. eg. An array with 6 columns and 4 rows
    can be reduced to have 6,3,2 or 1 columns and 4,2 or 1 rows.
    example usages:
    https://scipy-cookbook.readthedocs.io/items/Rebinning.html
    >>> a=rand(6,4); b=rebin(a,3,2)
    >>> a=rand(6); b=rebin(a,2)
    '''
    shape = a.shape
    lenShape = len(shape)
    factor = np.asarray(shape) // np.asarray(args)
    evList = ['a.reshape('] + \
             ['args[%d],factor[%d],' % (i, i) for i in range(lenShape)] + \
             [')'] + ['.sum(%d)' % (i + 1) for i in range(lenShape)] + \
             ['/factor[%d]' % i for i in range(lenShape - 1)]
    ta = eval(''.join(evList))
    return ta.astype(np.uint32), np.indices([s + 1 for s in ta.shape])


def plot_badpix_3d(data, definitions, title=None, rebin_fac=2, azim=22.5):
    od = data
    d, dims = rebin(od.astype(np.uint32), od.shape[0] // rebin_fac, od.shape[1] // rebin_fac, od.shape[2])
    xx, yy, zz = dims
    voxels = d.astype(np.bool)
    colors = np.full(voxels.shape, '#FFFFFF')
    cols = definitions

    for k, c in cols.items():
        colors[d == k] = c[1]

    fig = plt.figure(figsize=(15, 10))
    ax = fig.gca(projection="3d")
    ax.voxels(xx*rebin_fac, yy*rebin_fac, zz, voxels, facecolors=colors)
    ax.view_init(elev=25., azim=azim)
    ax.set_xlabel("pixels")
    ax.set_ylabel("pixels")
    ax.set_zlabel("memory cell")
    ax.set_xlim(0, np.max(xx) * rebin_fac)
    ax.set_ylim(0, np.max(yy) * rebin_fac)
    ax.set_zlim(0, np.max(zz))

    for k, c in cols.items():
        ax.plot([-1,], [-1,], color=c[1], label=c[0])
    ax.legend()
    if title:
        ax.set_title(title)


def create_constant_overview(constant, name, cells, vmin=None, vmax=None,
                             entries=3, badpixels=None, gmap=None,
                             marker=None):
    """
    Create a step plot for constant data across memory cells for requested
    gain entries

    :param constant: dict with constants for each module.
    :param name: Name to be used for the x-axis
    :param cells: Number of memory cells
    :param vmin: plot minumim value boundaries
    :param vmax: plot maximum value boundaries
    :param entries: (int)number of gain entries.
        A tuple specifying the range can also be used.
    TODO: remove unused inputs from notebooks.
    :param out_folder: out_folder for showing .md table statistics
    :param infix: infix for the output png image
    :param badpixels: A list of 2 elements.
        badpixels[0] has the dict with badpixels constant for each module
        and badpixels[1] has the value to apply for bad pixels.
    :param gmap: A list with len equal to number of gain entires.
        if not supported, a default would be used for 3 entries.
        ['High gain', 'Medium gain', 'Low gain']
    :param marker: A list of line markers for each gain entry.
        default: ['']*entries
    :return:
    """
    if gmap is None:
        gmap = ['High gain', 'Medium gain', 'Low gain']
    if marker is None:
        marker = [''] * entries
    fig = plt.figure(figsize=(10, 5))
    ax = fig.add_subplot(111)
    for g in range(entries):
        table = []
        dbp = None

        for qm in constant.keys():
            if len(constant[qm].shape) == 4:
                d = constant[qm][..., g]
                if badpixels is not None and isinstance(badpixels, list):
                    dbp = np.copy(d)
                    dbp[badpixels[0][qm][..., g] > 0] = badpixels[1]

            else:
                # e.g. DSSC
                d = constant[qm]
                if badpixels is not None and isinstance(badpixels, list):
                    dbp = np.copy(d)
                    dbp[badpixels[0][qm] > 0] = badpixels[1]
            # TODO: check if not used to remove.
            table.append([name, qm, gmap[g], np.nanmean(d), np.nanmedian(d),
                          np.nanstd(d)])

            ax.step(np.arange(cells), np.nanmean(d, axis=(0, 1)),
                    label=gmap[g], color=f'C{g}', marker=marker[g])
            # Plotting good pixels only if bad-pixels were given
            if dbp is not None:
                ax.step(np.arange(cells), np.nanmean(dbp, axis=(0, 1)),
                        label=f'Good pixels {gmap[g]}', color=f'C{g}',
                        linestyle='--')
    ax.set_xlabel("Memory cell")
    ax.set_ylabel(name)
    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    ax.set_title(f"{name} Median per Cell")
    if vmin and vmax:
        ax.set_ylim(vmin, vmax)


def show_processed_modules(dinstance, constants, mnames, mode):
    """
    Show the status of the processed modules.
    Green: Processed. Gray: Not Processed. Red: No data available.
    :param dinstance: The detector instance (e.g. AGIPD1M1 or LPD1M1)
    :param constants: A dict of the plotted constants data.
        {"const_name":constant_data}. Can be None in case of position mode.
    :param mnames: A list of available module names.
    :param mode: String selecting on of the two modes of operation.
         "position": To just show the position of the processed modules.
         "processed": To show the modules successfully processed.
    :return
    """
    fig, ax = plt.subplots(1, figsize=(10, 10))
    ax.set_axis_off()
    counter = 0

    if 'AGIPD' in dinstance:

        ax.set_xlim(0, 90)
        ax.set_ylim(0, 75)
        asic_pos = 5
        q_st = 8
        # quad_pos = [Q1[x, y], Q2[x, y], Q3[x, y], Q4[x, y]]
        quad_pos = [[43, 66], [45, 34], [4, 32], [2, 64]]
        l_y = 6
        l_x = 5
        for iq, q_x in enumerate(quad_pos):
            for im in range(4):
                if mode == "processed":
                    color = 'gray'
                    ecolor = 'darkgreen'
                    if f'Q{iq+1}M{im+1}' in mnames:
                        color = 'green'
                        if 'Noise' not in constants.keys() or \
                                np.nanmean(constants['Noise'][counter, :, :, :, 0]) == 0:  # noqa
                            color = 'red'
                        counter += 1
                else:
                    color = 'lightblue'
                    ecolor = 'c'
                x = q_x[0]
                for _ in range(8):
                    # adding modules patches
                    ax.add_patch(Rectangle((x, q_x[1] - q_st * im), 5, 6,
                                           linewidth=2, edgecolor=ecolor,
                                           facecolor=color, fill=True))
                    if mode == "position":
                        if f"Q{iq+1}M{im+1}" in mnames:
                            ax.add_patch(
                                Rectangle((x, q_x[1] - q_st * im),
                                          l_x, l_y / 2, linewidth=2,
                                          edgecolor='c',
                                          facecolor='sandybrown',
                                          fill=True))
                            ax.add_patch(
                                Rectangle((x, (q_x[1] - q_st * im + 3)),
                                          l_x, l_y / 2, linewidth=2,
                                          edgecolor='c',
                                          facecolor='sandybrown',
                                          fill=True))
                    x += asic_pos
                if mode == "position":
                    if f"Q{iq+1}M{im+1}" in mnames:
                        # Change Text for current processed module.
                        ax.text(q_x[0] + 13, q_x[1] - q_st * im + 1.5,
                                f'Q{iq+1}M{im+1}', fontsize=28,
                                color='mediumblue', weight='bold')
                    else:
                        ax.text(q_x[0] + 14, q_x[1] - q_st * im + 1.5,
                                f'Q{iq+1}M{im+1}', fontsize=25, color='k')
                elif mode == "processed":
                    ax.text(q_x[0] + 14.5, q_x[1] - q_st * im + 1.5,
                            f'Q{iq+1}M{im+1}', fontsize=24, color='gold')
    elif 'LPD' in dinstance:

        ax.set_xlim(0, 97)
        ax.set_ylim(0, 97)
        q_poses = np.array([[51, 47], [47, 1], [1, 5], [5, 51]])
        m_poses = np.array([[22.5, 20.5], [22.5, 0.5],
                            [0.5, 0.5], [0.5, 20.5]])

        for iq, q_pos in enumerate(q_poses):
            ax.add_patch(Rectangle(q_pos, 45, 45, linewidth=2,
                                   edgecolor='darkgreen',
                                   facecolor='cornsilk', fill=True))
            ax.text(q_pos[0] + 20, q_pos[1] + 41.5, f'Q{iq+1}', fontsize=22)

            for im, m_pos in enumerate(m_poses):
                if mode == "processed":
                    color = 'gray'
                    ecolor = 'darkgreen'
                    if f'Q{iq+1}M{im+1}' in mnames:
                        color = 'green'
                        if 'Noise' not in constants.keys() or \
                                np.nanmean(constants['Noise'][counter, :, :, :, 0]) == 0:  # noqa
                            color = 'red'
                            counter += 1
                    pos = q_pos + m_pos + np.array([5, 8])
                    ax.text(pos[0], pos[1], f'Q{iq+1}M{im+1}',
                            fontsize=24, color='gold')
                elif mode == "position":
                    color = 'lightblue'
                    ecolor = 'c'
                # adding modules patches
                ax.add_patch(Rectangle(q_pos + m_pos, 22, 20, linewidth=3,
                                       edgecolor=ecolor, facecolor=color,
                                       fill=True))
                if mode == "position":
                    if f"Q{iq+1}M{im+1}" in mnames:
                        for a_posx in range(2):
                            for a_posy in range(8):
                                a_pos = np.array([a_posx*11., a_posy*20/8.])
                                pos = q_pos + m_pos + a_pos
                                # Top Asics patches
                                ax.add_patch(Rectangle(q_pos+m_pos+a_pos, 11,
                                                       20/8., linewidth=1,
                                                       edgecolor='k',
                                                       facecolor='sandybrown',
                                                       fill=True))
                                if a_posx == 0:
                                    label = str(a_posy + 9)
                                else:
                                    label = str(-a_posy + (a_posx * 8))
                                ax.text(pos[0] + 4, pos[1] + 0.3, label,
                                        fontsize=14)
                    else:
                        pos = q_pos + m_pos + np.array([5, 8])
                        ax.text(pos[0], pos[1], f'Q{iq+1}M{im+1}',
                                fontsize=22, color='k')

            if mode == "position":
                ax.add_patch(Rectangle([65, 93], 30, 4, linewidth=1,
                                       edgecolor='k', facecolor='sandybrown',
                                       fill=True))
                ax.text(52, 94, 'ASICs:', fontsize=22, color='k')

                for i_pos in range(8):
                    pos = np.array([65, 93]) + np.array([i_pos*30/8.+0.3, 0.3])  # noqa

                    ax.add_patch(Rectangle(pos, 24/8., 3.4, linewidth=1,
                                           edgecolor='k',
                                           facecolor='sandybrown',
                                           fill=True))

                    ax.text(pos[0]+0.5, pos[1]+0.5, f'{i_pos+1}',
                            fontsize=18, color='k')

    elif 'DSSC' in dinstance:
        ax.set_xlim(0, 85)
        ax.set_ylim(0, 79)
        asic_pos = 20
        q_st = 8

        # quad_pos = [Q1[x, y], Q2[x, y], Q3[x, y], Q4[x, y]]
        quad_pos = [[45, 64], [43, 24], [0, 28], [0, 66]]
        mod_ind = [[4, 3, 2, 1], [4, 3, 2, 1], [1, 2, 3, 4], [1, 2, 3, 4]]

        for iq, q_x in enumerate(quad_pos):
            for im in range(4):
                if mode == "processed":
                    color = 'gray'
                    ecolor = 'darkgreen'
                    if f'Q{iq+1}M{mod_ind[iq][im]}' in mnames:
                        color = 'green'
                        if 'Noise' not in constants.keys() or \
                                np.nanmean(constants['Noise'][counter, :, :, :]) == 0:  # noqa
                            color = 'red'
                        counter += 1
                else:
                    color = 'lightblue'
                    ecolor = 'c'
                x = q_x[0]
                for _ in range(2):
                    # adding modules patches
                    ax.add_patch(Rectangle((x, q_x[1] - q_st * im), 19, 10,
                                           linewidth=2, edgecolor=ecolor,
                                           facecolor=color, fill=True))
                    if mode == "position":
                        # Check if it is the processed module to highlight.
                        if f"Q{iq+1}M{mod_ind[iq][im]}" in mnames:
                            ax.add_patch(Rectangle((x, q_x[1] - q_st * im),
                                                   19, 10, linewidth=2,
                                                   edgecolor='c',
                                                   facecolor="sandybrown",
                                                   fill=True))

                        ax.text(q_x[0] + 2, q_x[1] - q_st * im + 4.5,
                                'T1                              T2', fontsize=18,  # noqa
                                color='grey')
                    x += asic_pos
                if mode == "position":
                    # Check if it is the processed module to highlight.
                    if f'Q{iq+1}M{mod_ind[iq][im]}' in mnames:
                        # Change Text for current processed module.
                        ax.text(q_x[0] + 12.7, q_x[1] - q_st * im + 4.5,
                                f'Q{iq + 1} M{mod_ind[iq][im]}', fontsize=26,
                                color='mediumblue', weight='bold')
                    else:
                        ax.text(q_x[0] + 13.5, q_x[1] - q_st * im + 4.5,
                                f'Q{iq+1} M{mod_ind[iq][im]}',
                                fontsize=25, color='k')
                elif mode == "processed":
                    ax.text(q_x[0] + 13.5, q_x[1] - q_st * im + 4.5,
                            f'Q{iq+1} M{mod_ind[iq][im]}', fontsize=24,
                            color='gold')
                    ax.text(q_x[0] + 2, q_x[1] - q_st * im + 4.5,
                            'T1                              T2', fontsize=18,
                            color='y')
    if mode == "processed":
        _ = ax.legend(handles=[Patch(facecolor='red', label='No data'),
                               Patch(facecolor='gray', label='Not processed'),
                               Patch(facecolor='green', label='Processed')],
                      loc='outside-top', ncol=3, bbox_to_anchor=(0.1, 0.25, 0.7, 0.8))
    plt.show()
