from collections import OrderedDict
import datetime
from glob import glob
import json
from os import environ, listdir, path, stat
from os.path import isfile, splitext
from queue import Queue
import re
from time import sleep
from urllib.parse import urljoin

import dateutil.parser

import ipykernel
from metadata_client.metadata_client import MetadataClient
from notebook.notebookapp import list_running_servers
import numpy as np
import requests

from .mdc_config import MDC_config
from .ana_tools import save_dict_to_hdf5

def parse_runs(runs, return_type=str):
    pruns = []
    if isinstance(runs, str):
        for rcomp in runs.split(","):
            if "-" in rcomp:
                start, end = rcomp.split("-")
                pruns += list(range(int(start), int(end)))
            else:
                pruns += [int(rcomp)]
    elif isinstance(runs, (list, tuple)):
        pruns = runs
    else:
        pruns = [runs, ]

    if return_type is str:
        return ["r{:04d}".format(r) for r in pruns]
    else:
        return pruns


def run_prop_seq_from_path(filename):
    run = re.findall(r'.*r([0-9]{4}).*', filename)
    run = run[0] if len(run) else None
    proposal = re.findall(r'.*p([0-9]{6}).*', filename)
    proposal = proposal[0] if len(proposal) else None
    sequence = re.findall(r'.*S([0-9]{5}).*', filename)
    sequence = sequence[0] if len(sequence) else None
    return run, proposal, sequence


def map_modules_from_folder(in_folder, run, path_template, karabo_da,
                            sequences=None):
    """
    Prepare queues of files to process.
    Queues are stored in dictionary with module name Q{}M{} as a key

    :param in_folder: Input folder with raw data
    :param run: Run number
    :param path_template: Template for file name e.g. `RAW-R{:04d}-{}-S{:05d}.h5`
    :param karabo_da: List of data aggregators e.g. [AGIPD00, AGIPD01]
    :param sequences: List of sequences to be considered
    :return: Dictionary of queues of files, dictionary of module indexes,
    total number of sequences, dictionary of number of sequences per module
    """
    module_files = OrderedDict()
    mod_ids = OrderedDict()
    total_sequences = 0
    total_file_size = 0
    sequences_qm = {}
    for inset in karabo_da:
        module_idx = int(inset[-2:])
        name = f"Q{module_idx // 4 + 1}M{module_idx % 4 + 1}"
        module_files[name] = Queue()
        sequences_qm[name] = 0
        mod_ids[name] = module_idx
        if sequences is None:
            fname = path_template.format(run, inset, 0).replace("S00000", "S*")
            abs_fname = "{}/r{:04d}/{}".format(in_folder, run, fname)

            for filename in glob(abs_fname):
                module_files[name].put(filename)
                total_sequences += 1
                sequences_qm[name] += 1
                total_file_size += path.getsize(filename)
        else:
            for sequence in sequences:
                fname = path_template.format(run, inset, sequence)
                abs_fname = "{}/r{:04d}/{}".format(in_folder, run, fname)
                if not isfile(abs_fname):
                    continue

                module_files[name].put(abs_fname)
                total_sequences += 1
                sequences_qm[name] += 1
                total_file_size += path.getsize(abs_fname)

    return (module_files, mod_ids, total_sequences,
            sequences_qm, total_file_size)


def map_gain_stages(in_folder, runs, path_template, karabo_da, sequences=None):
    """
    Prepare queues of files to process.
    Queues are stored in dictionary with module name Q{}M{}
    and gain name as a keys
    :param in_folder: Input folder with raw data
    :param runs: Dictionary of runs with key naming the gain stages
    :param path_template: Template for file name e.g. `RAW-R{:04d}-{}-S{:05d}.h5`
    :param karabo_da: List of data aggregators e.g. [AGIPD00, AGIPD01]
    :param sequences: List of sequences to be considered
    :return: Dictionary of queues of files,
    total number of sequences
    """
    total_sequences = 0
    total_file_size = 0
    gain_mapped_files = OrderedDict()
    for gain, run in runs.items():
        mapped_files, _, seq, _, fs = map_modules_from_folder(in_folder, run,
                                                              path_template,
                                                              karabo_da,
                                                              sequences)

        total_sequences += seq
        total_file_size += fs
        gain_mapped_files[gain] = mapped_files
    return gain_mapped_files, total_sequences, total_file_size / 1e9


def map_modules_from_files(filelist, file_inset, quadrants, modules_per_quad):
    total_sequences = 0
    total_file_size = 0
    module_files = {}
    mod_ids = {}
    for quadrant in range(0, quadrants):
        for module in range(0, modules_per_quad):
            name = "Q{}M{}".format(quadrant + 1, module + 1)
            module_files[name] = Queue()
            num = quadrant * 4 + module
            mod_ids[name] = num
            file_infix = "{}{:02d}".format(file_inset, num)
            for file in filelist:
                if file_infix in file:
                    module_files[name].put(file)
                    total_sequences += 1
                    total_file_size += path.getsize(file)

    return module_files, mod_ids, total_sequences, total_file_size


def gain_map_files(in_folder, runs, sequences, file_inset, quadrants,
                   mods_per_quad):
    total_sequences = 0
    total_file_size = 0
    gain_mapped_files = OrderedDict()
    for gain, run in runs.items():
        ginfolder = "{}/{}".format(in_folder, run)
        dirlist = listdir(ginfolder)
        file_list = []
        for entry in dirlist:
            # only h5 file
            abs_entry = "{}/{}".format(ginfolder, entry)
            if path.isfile(abs_entry) and path.splitext(abs_entry)[1] == ".h5":

                if sequences is None:
                    file_list.append(abs_entry)
                else:
                    for seq in sequences:
                        if "{:05d}.h5".format(seq) in abs_entry:
                            file_list.append(path.abspath(abs_entry))

        mapped_files, mod_ids, seq, fs = map_modules_from_files(file_list,
                                                                file_inset,
                                                                quadrants,
                                                                mods_per_quad)
        total_sequences += seq
        total_file_size += fs
        gain_mapped_files[gain] = mapped_files
    return gain_mapped_files, total_sequences, total_file_size / 1e9


def get_notebook_name():
    """
    Return the full path of the jupyter notebook.
    """
    try:
        kernel_id = re.search('kernel-(.*).json',
                              ipykernel.connect.get_connection_file()).group(1)
        servers = list_running_servers()
        for ss in servers:
            response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                    params={'token': ss.get('token', '')})
            for nn in json.loads(response.text):
                if nn['kernel']['id'] == kernel_id:
                    relative_path = nn['notebook']['path']
                    return relative_path
    except:
        return environ.get("CAL_NOTEBOOK_NAME", "Unknown Notebook")


def get_run_info(proposal, run):
    """
    Return information about run from the MDC

    :param proposal: proposal number
    :param run: run number
    :return: dictionary with run information
    """

    mdconf = MDC_config.mdconf
    mdc = MetadataClient(client_id=mdconf['user-id'],
                         client_secret=mdconf['user-secret'],
                         user_email=mdconf['user-email'],
                         token_url=mdconf['token-url'],
                         refresh_url=mdconf['refresh-url'],
                         auth_url=mdconf['auth-url'],
                         scope=mdconf['scope'],
                         base_api_url=mdconf['base-api-url'])

    runs = mdc.get_proposal_runs(mdc, proposal_number=proposal,
                                 run_number=run)
    run_id = runs['data']['runs'][0]['id']

    resp = mdc.get_run_by_id_api(run_id)
    return resp.json()


def get_dir_creation_date(directory, run, tsdir=False, verbosity=0):
    """
    Return run starting time from the MDC.
    If not succeeded, return modification time of oldest file.h5
    in [directory]/[run]04.

    :param directory: path to directory which contains runs
    :param run: run number
    :param tsdir: to get modification time of [directory]/[run]04.
    :param verbosity: Level of verbosity (0 - silent)
    :return: (datetime) modification time
    """

    try:
        path_list = list(filter(None, directory.strip('/').split('/')))
        proposal = int(path_list[-2][1:])
        run_info = get_run_info(proposal, run)
        return dateutil.parser.parse(run_info['begin_at'])
    except Exception as e:
        if verbosity > 0:
            print(e)

    ntries = 100
    while ntries > 0:
        try:
            if tsdir:
                creation_time = stat("{}/r{:04d}".format(directory, run)).st_mtime
            else:
                rfiles = glob("{}/r{:04d}/*.h5".format(directory, run))
                rfiles.sort(key=path.getmtime)
                creation_time = stat(rfiles[0]).st_mtime

            creation_time = datetime.datetime.fromtimestamp(creation_time)
            return creation_time
        except:  # catch stale file handle errors etc and try again
            ntries -= 1


def save_const_to_h5(metadata, out_folder):
    """
    Save constant in h5 file with it's metadata
    (e.g. db_module, condition, creation_time)

    :param metadata: Metadata
    :param out_folder: path to output folder
    """

    dpar = {}
    for parm in metadata.detector_condition.parameters:
        dpar[parm.name] = {'lower_deviation_value': parm.lower_deviation,
                           'upper_deviation_value': parm.upper_deviation,
                           'value': parm.value,
                           'flg_logarithmic': parm.logarithmic}

    creation_time = metadata.calibration_constant_version.begin_at
    raw_data = metadata.calibration_constant_version.raw_data_location
    db_module = metadata.calibration_constant_version.device_name
    constant_name = metadata.calibration_constant.__class__.__name__

    data_to_store = {}
    data_to_store['condition'] = dpar
    data_to_store['db_module'] = db_module
    data_to_store['constant'] = constant_name
    data_to_store['data'] = metadata.calibration_constant.data
    data_to_store['creation_time'] = creation_time
    data_to_store['file_loc'] = raw_data

    ofile = "{}/const_{}_{}.h5".format(out_folder, constant_name, db_module)
    if isfile(ofile):
        print('File {} already exists and will be overwritten'.format(ofile))
    save_dict_to_hdf5(data_to_store, ofile)


def get_random_db_interface(cal_db_interface):
    """
    Return interface to calibration DB with random (with given range) port
    """
    if "#" in cal_db_interface:
        prot, serv, ran = cal_db_interface.split(":")
        r1, r2 = ran.split("#")
        return ":".join([prot, serv, str(np.random.randint(int(r1), int(r2)))])
    return cal_db_interface


already_printed = {}


def get_from_db(device, constant, condition, empty_constant,
                cal_db_interface, creation_time=None,
                verbosity=1, timeout=30000, ntries=7, meta_only=True,
                version_info=False, doraise=False):
    """
    Return calibration constants and metadata requested from CalDB

    :param device: Instance of detector
    :param constant: Calibration constant known for given detector
    :param condition: Calibration condition
    :param empty_constant: Constant to be returned in case of failure.
    :param cal_db_interface: Interface string, e.g. "tcp://max-exfl016:8015"
    :param creation_time: Latest time for constant to be created
    :param verbosity: Level of verbosity (0 - silent)
    :param timeout: Timeout for zmq request
    ntries is set to 7 so that if the timeout started at 30s last timeout
    will be ~ 1h.
    :param ntries: number of tries to contact the database
    :param meta_only: (bool) Retrieve only metadata via ZMQ. Constants are
        taken directly from the h5 file on maxwell.
    :param version_info: (bool) to show the info for the retrieved Constant
    :param doraise: (bool) if True raise errors during communication with DB.
    :return: Calibration constant, metadata
    """
    from iCalibrationDB import ConstantMetaData, Versions
    import zmq

    if version_info:
        meta_only = False

    if device:
        metadata = ConstantMetaData()
        metadata.calibration_constant = constant
        metadata.detector_condition = condition
        when = None
        if creation_time is None:
            metadata.calibration_constant_version = Versions.Now(
                                                    device=device)
        else:
            if hasattr(creation_time, 'isoformat'):
                when = creation_time.isoformat()

            metadata.calibration_constant_version = Versions.Timespan(
                                                    device=device,
                                                    start=creation_time)
        while ntries > 0:

            this_interface = get_random_db_interface(cal_db_interface)
            try:
                r = metadata.retrieve(this_interface, timeout=timeout,
                                      when=when, meta_only=meta_only,
                                      version_info=version_info)
                if version_info:
                    return r
                break
            except zmq.error.Again:
                ntries -= 1
                timeout *= 2
                sleep(np.random.randint(30))
                if ntries == 0 and doraise:
                    raise
            except Exception as e:
                if verbosity > 0:
                    print(e)
                if 'missing_token' in str(e):
                    ntries -= 1
                else:
                    ntries = 0
                if ntries == 0 and doraise:
                    raise RuntimeError(f'{e}')


        if ntries > 0:
            if verbosity > 0:
                if constant.name not in already_printed or verbosity > 1:
                    already_printed[constant.name] = True
                    begin_at = metadata.calibration_constant_version.begin_at
                    print(f"Retrieved {constant.name} "
                          f"with creation time: {begin_at}")
            return constant.data, metadata
        else:
            return empty_constant, metadata
    else:
        return empty_constant, None


def send_to_db(device, constant, condition, file_loc,
               cal_db_interface, creation_time=None,
               verbosity=1, timeout=30000, ntries=7, doraise=False):
    """
    Return calibration constants and metadata requested from CalDB

    :param device: Instance of detector
    :param constant: Calibration constant known for given detector
    :param condition: Calibration condition
    :param file_loc: Location of raw data.
    :param cal_db_interface: Interface string, e.g. "tcp://max-exfl016:8015"
    :param creation_time: Latest time for constant to be created
    :param verbosity: Level of verbosity (0 - silent)
    :param timeout: Timeout for zmq request
    :param ntries: number of tries to contact the database,
    ntries is set to 7 so that if the timeout started at 30s last timeout
    will be ~ 1h.
    :param doraise: (bool) if True raise errors during communication with DB
    """
    from iCalibrationDB import ConstantMetaData, Versions
    import zmq

    if device:
        metadata = ConstantMetaData()
        metadata.calibration_constant = constant
        metadata.detector_condition = condition
        if creation_time is None:
            metadata.calibration_constant_version = Versions.Now(
                device=device)
        else:
            metadata.calibration_constant_version = Versions.Timespan(
                device=device,
                start=creation_time)

        metadata.calibration_constant_version.raw_data_location = file_loc

        while ntries > 0:

            this_interface = get_random_db_interface(cal_db_interface)
            try:
                r = metadata.send(this_interface, timeout=timeout)
                break
            except zmq.error.Again:
                ntries -= 1
                timeout *= 2
                sleep(np.random.randint(30))
                if ntries == 0 and doraise:
                    raise
            except Exception as e:
                if verbosity > 0:
                    print(e)
                if 'missing_token' in str(e):
                    ntries -= 1
                else:
                    ntries = 0
                if ntries == 0 and doraise:
                    raise RuntimeError(f'{e}')

        if ntries > 0:
            if verbosity > 0:
                if constant.name not in already_printed or verbosity > 1:
                    already_printed[constant.name] = True
                    begin_at = metadata.calibration_constant_version.begin_at
                    print("{} was sent on: {}".format(constant.name,
                                                      begin_at))


def get_constant_from_db(device, constant, condition, empty_constant,
                         cal_db_interface, creation_time=None,
                         print_once=True, timeout=30000, ntries=120,
                         meta_only=True):
    """
    Return calibration constants requested from CalDB
    """
    data, _ = get_from_db(device, constant, condition, empty_constant,
                          cal_db_interface, creation_time,
                          int(print_once), timeout, ntries, meta_only)
    return data


def get_constant_from_db_and_time(device, constant, condition, empty_constant,
                                  cal_db_interface, creation_time=None,
                                  print_once=True, timeout=30000, ntries=120):
    """
    Return calibration constants requested from CalDB, alongside injection
    time
    """
    data, m = get_from_db(device, constant, condition, empty_constant,
                          cal_db_interface, creation_time,
                          int(print_once), timeout, ntries)
    if m:
        if m.comm_db_success:
            return data, m.calibration_constant_version.begin_at
        else:
            # retun none for injection time if communication with db failed.
            # reasons (no constant or condition found,
            # or network problem)
            return data, None
    else:
        return data, None
